from datetime import datetime, timezone
from logging import Logger
from typing import NamedTuple, cast

from unique_follow_up_questions.follow_up_postprocessor import (
    FollowUpPostprocessor,
)
from unique_internal_search.uploaded_search.config import (
    UploadedSearchConfig,
)
from unique_internal_search.uploaded_search.service import (
    UploadedSearchTool,
)
from unique_stock_ticker.stock_ticker_postprocessor import (
    StockTickerPostprocessor,
)
from unique_toolkit import LanguageModelService, get_async_openai_client
from unique_toolkit.agentic.debug_info_manager.debug_info_manager import (
    DebugInfoManager,
)
from unique_toolkit.agentic.evaluation.evaluation_manager import EvaluationManager
from unique_toolkit.agentic.evaluation.hallucination.hallucination_evaluation import (
    HallucinationEvaluation,
)
from unique_toolkit.agentic.history_manager import (
    history_manager as history_manager_module,
)
from unique_toolkit.agentic.history_manager.history_manager import (
    HistoryManager,
    HistoryManagerConfig,
)
from unique_toolkit.agentic.message_log_manager.service import MessageStepLogger
from unique_toolkit.agentic.postprocessor.postprocessor_manager import (
    PostprocessorManager,
)
from unique_toolkit.agentic.reference_manager.reference_manager import ReferenceManager
from unique_toolkit.agentic.responses_api import (
    DisplayCodeInterpreterFilesPostProcessor,
    ShowExecutedCodePostprocessor,
)
from unique_toolkit.agentic.thinking_manager.thinking_manager import (
    ThinkingManager,
    ThinkingManagerConfig,
)
from unique_toolkit.agentic.tools.a2a import (
    A2AManager,
    ExtendedSubAgentToolConfig,
    SubAgentDisplaySpec,
    SubAgentEvaluationService,
    SubAgentEvaluationSpec,
    SubAgentReferencesPostprocessor,
    SubAgentResponsesDisplayPostprocessor,
    SubAgentResponsesPostprocessorConfig,
    SubAgentResponseWatcher,
)
from unique_toolkit.agentic.tools.config import ToolBuildConfig
from unique_toolkit.agentic.tools.mcp.manager import MCPManager
from unique_toolkit.agentic.tools.openai_builtin.base import OpenAIBuiltInToolName
from unique_toolkit.agentic.tools.tool_manager import (
    OpenAIBuiltInToolManager,
    ResponsesApiToolManager,
    ToolManager,
    ToolManagerConfig,
)
from unique_toolkit.agentic.tools.tool_progress_reporter import ToolProgressReporter
from unique_toolkit.app.schemas import ChatEvent, McpServer
from unique_toolkit.chat.service import ChatService
from unique_toolkit.content import Content
from unique_toolkit.content.service import ContentService
from unique_toolkit.protocols.support import ResponsesSupportCompleteWithReferences

from unique_orchestrator.config import UniqueAIConfig
from unique_orchestrator.unique_ai import UniqueAI, UniqueAIResponsesApi


async def build_unique_ai(
    event: ChatEvent,
    logger: Logger,
    config: UniqueAIConfig,
    debug_info_manager: DebugInfoManager,
) -> UniqueAI | UniqueAIResponsesApi:
    common_components = _build_common(event, logger, config)

    if config.agent.experimental.responses_api_config.use_responses_api:
        return await _build_responses(
            event=event,
            logger=logger,
            config=config,
            debug_info_manager=debug_info_manager,
            common_components=common_components,
        )
    else:
        return _build_completions(
            event=event,
            logger=logger,
            config=config,
            debug_info_manager=debug_info_manager,
            common_components=common_components,
        )


class _CommonComponents(NamedTuple):
    chat_service: ChatService
    content_service: ContentService
    uploaded_documents: list[Content]
    thinking_manager: ThinkingManager
    reference_manager: ReferenceManager
    history_manager: HistoryManager
    evaluation_manager: EvaluationManager
    postprocessor_manager: PostprocessorManager
    message_step_logger: MessageStepLogger
    response_watcher: SubAgentResponseWatcher
    # Tool Manager Components
    tool_progress_reporter: ToolProgressReporter
    tool_manager_config: ToolManagerConfig
    mcp_manager: MCPManager
    a2a_manager: A2AManager
    mcp_servers: list[McpServer]


def _build_common(
    event: ChatEvent,
    logger: Logger,
    config: UniqueAIConfig,
) -> _CommonComponents:
    chat_service = ChatService(event)

    content_service = ContentService.from_event(event)

    uploaded_documents = content_service.get_documents_uploaded_to_chat()

    response_watcher = SubAgentResponseWatcher()

    tool_progress_reporter = ToolProgressReporter(
        chat_service=chat_service,
        config=config.agent.services.tool_progress_reporter_config,
    )
    thinking_manager_config = ThinkingManagerConfig(
        thinking_steps_display=config.agent.experimental.thinking_steps_display
    )
    thinking_manager = ThinkingManager(
        logger=logger,
        config=thinking_manager_config,
        tool_progress_reporter=tool_progress_reporter,
        chat_service=chat_service,
    )

    reference_manager = ReferenceManager()

    history_manager_config = HistoryManagerConfig(
        experimental_features=history_manager_module.ExperimentalFeatures(),
        percent_of_max_tokens_for_history=config.agent.input_token_distribution.percent_for_history,
        language_model=config.space.language_model,
        uploaded_content_config=config.agent.services.uploaded_content_config,
    )
    history_manager = HistoryManager(
        logger,
        event,
        history_manager_config,
        config.space.language_model,
        reference_manager,
    )

    evaluation_manager = EvaluationManager(logger=logger, chat_service=chat_service)
    if config.agent.services.evaluation_config:
        evaluation_manager.add_evaluation(
            HallucinationEvaluation(
                config.agent.services.evaluation_config.hallucination_config,
                event,
                reference_manager,
            )
        )

    mcp_manager = MCPManager(
        mcp_servers=event.payload.mcp_servers,
        event=event,
        tool_progress_reporter=tool_progress_reporter,
    )
    a2a_manager = A2AManager(
        logger=logger,
        tool_progress_reporter=tool_progress_reporter,
        response_watcher=response_watcher,
    )

    tool_manager_config = ToolManagerConfig(
        tools=config.space.tools,
        max_tool_calls=config.agent.experimental.loop_configuration.max_tool_calls_per_iteration,
    )

    postprocessor_manager = PostprocessorManager(
        logger=logger,
        chat_service=chat_service,
    )

    if config.agent.services.stock_ticker_config is not None:
        postprocessor_manager.add_postprocessor(
            StockTickerPostprocessor(
                config=config.agent.services.stock_ticker_config,
                event=event,
            )
        )

    if (
        config.agent.services.follow_up_questions_config
        and config.agent.services.follow_up_questions_config.number_of_questions > 0
    ):
        # Should run last to make sure the follow up questions are displayed last.
        postprocessor_manager.set_last_postprocessor(
            FollowUpPostprocessor(
                logger=logger,
                config=config.agent.services.follow_up_questions_config,
                event=event,
                historyManager=history_manager,
                llm_service=LanguageModelService.from_event(event),
            )
        )

    return _CommonComponents(
        chat_service=chat_service,
        content_service=content_service,
        uploaded_documents=uploaded_documents,
        thinking_manager=thinking_manager,
        reference_manager=reference_manager,
        history_manager=history_manager,
        evaluation_manager=evaluation_manager,
        tool_progress_reporter=tool_progress_reporter,
        mcp_manager=mcp_manager,
        a2a_manager=a2a_manager,
        tool_manager_config=tool_manager_config,
        mcp_servers=event.payload.mcp_servers,
        postprocessor_manager=postprocessor_manager,
        response_watcher=response_watcher,
        message_step_logger=MessageStepLogger(chat_service),
    )


async def _build_responses(
    event: ChatEvent,
    logger: Logger,
    config: UniqueAIConfig,
    common_components: _CommonComponents,
    debug_info_manager: DebugInfoManager,
) -> UniqueAIResponsesApi:
    client = get_async_openai_client().copy(
        default_headers={
            "x-model": config.space.language_model.name,
            "x-user-id": event.user_id,
            "x-company-id": event.company_id,
            "x-assistant-id": event.payload.assistant_id,
            "x-chat-id": event.payload.chat_id,
        }
    )

    assert config.agent.experimental.responses_api_config is not None

    code_interpreter_config = (
        config.agent.experimental.responses_api_config.code_interpreter
    )
    postprocessor_manager = common_components.postprocessor_manager
    tool_names = [tool.name for tool in config.space.tools]

    if code_interpreter_config is not None:
        if OpenAIBuiltInToolName.CODE_INTERPRETER not in tool_names:
            logger.info("Automatically adding code interpreter to the tools")
            config = config.model_copy(deep=True)
            config.space.tools.append(
                ToolBuildConfig(
                    name=OpenAIBuiltInToolName.CODE_INTERPRETER,
                    configuration=code_interpreter_config.tool_config,
                )
            )
            common_components.tool_manager_config.tools = config.space.tools

        if code_interpreter_config.executed_code_display_config is not None:
            postprocessor_manager.add_postprocessor(
                ShowExecutedCodePostprocessor(
                    config=code_interpreter_config.executed_code_display_config
                )
            )

        postprocessor_manager.add_postprocessor(
            DisplayCodeInterpreterFilesPostProcessor(
                client=client,
                content_service=common_components.content_service,
                config=code_interpreter_config.generated_files_config,
                user_id=event.user_id,
                company_id=event.company_id,
                chat_id=event.payload.chat_id,
                chat_service=common_components.chat_service,
            )
        )

    builtin_tool_manager = await OpenAIBuiltInToolManager.build_manager(
        uploaded_files=common_components.uploaded_documents,
        content_service=common_components.content_service,
        user_id=event.user_id,
        company_id=event.company_id,
        chat_id=event.payload.chat_id,
        client=client,
        tool_configs=config.space.tools,
    )

    tool_manager = ResponsesApiToolManager(
        logger=logger,
        config=common_components.tool_manager_config,
        event=event,
        tool_progress_reporter=common_components.tool_progress_reporter,
        mcp_manager=common_components.mcp_manager,
        a2a_manager=common_components.a2a_manager,
        builtin_tool_manager=builtin_tool_manager,
    )

    postprocessor_manager = common_components.postprocessor_manager

    class ResponsesStreamingHandler(ResponsesSupportCompleteWithReferences):
        def complete_with_references(self, *args, **kwargs):
            return common_components.chat_service.complete_responses_with_references(
                *args, **kwargs
            )

        async def complete_with_references_async(self, *args, **kwargs):
            return await common_components.chat_service.complete_responses_with_references_async(
                *args, **kwargs
            )

    streaming_handler = ResponsesStreamingHandler()

    _add_sub_agents_postprocessor(
        postprocessor_manager=postprocessor_manager,
        tool_manager=tool_manager,
        config=config,
        response_watcher=common_components.response_watcher,
    )
    _add_sub_agents_evaluation(
        evaluation_manager=common_components.evaluation_manager,
        tool_manager=tool_manager,
        config=config,
        event=event,
        response_watcher=common_components.response_watcher,
    )

    return UniqueAIResponsesApi(
        event=event,
        config=config,
        logger=logger,
        chat_service=common_components.chat_service,
        content_service=common_components.content_service,
        tool_manager=tool_manager,
        thinking_manager=common_components.thinking_manager,
        streaming_handler=streaming_handler,
        history_manager=common_components.history_manager,
        reference_manager=common_components.reference_manager,
        evaluation_manager=common_components.evaluation_manager,
        postprocessor_manager=postprocessor_manager,
        debug_info_manager=debug_info_manager,
        message_step_logger=common_components.message_step_logger,
        mcp_servers=event.payload.mcp_servers,
    )


def _build_completions(
    event: ChatEvent,
    logger: Logger,
    config: UniqueAIConfig,
    common_components: _CommonComponents,
    debug_info_manager: DebugInfoManager,
) -> UniqueAI:
    # Uploaded content behavior is always to force uploaded search tool:
    # 1. Add it to forced tools if there are tool choices.
    # 2. Simply force it if there are no tool choices.
    # 3. Not available if not uploaded documents.
    now = datetime.now(timezone.utc)
    UPLOADED_DOCUMENTS_VALID = [
        doc
        for doc in common_components.uploaded_documents
        if doc.expired_at is None or doc.expired_at > now
    ]
    UPLOADED_DOCUMENTS_EXPIRED = [
        doc
        for doc in common_components.uploaded_documents
        if doc.expired_at is not None and doc.expired_at <= now
    ]
    TOOL_CHOICES = len(event.payload.tool_choices) > 0

    if UPLOADED_DOCUMENTS_EXPIRED:
        logger.info(
            f"Number of expired uploaded documents: {len(UPLOADED_DOCUMENTS_EXPIRED)}"
        )

    if UPLOADED_DOCUMENTS_VALID:
        logger.info(
            f"Number of valid uploaded documents: {len(UPLOADED_DOCUMENTS_VALID)}"
        )
        common_components.tool_manager_config.tools.append(
            ToolBuildConfig(
                name=UploadedSearchTool.name,
                display_name=UploadedSearchTool.name,
                configuration=UploadedSearchConfig(),
            )
        )
    if TOOL_CHOICES and UPLOADED_DOCUMENTS_VALID:
        event.payload.tool_choices.append(str(UploadedSearchTool.name))

    tool_manager = ToolManager(
        logger=logger,
        config=common_components.tool_manager_config,
        event=event,
        tool_progress_reporter=common_components.tool_progress_reporter,
        mcp_manager=common_components.mcp_manager,
        a2a_manager=common_components.a2a_manager,
    )
    if not TOOL_CHOICES and UPLOADED_DOCUMENTS_VALID:
        tool_manager.add_forced_tool(UploadedSearchTool.name)

    postprocessor_manager = common_components.postprocessor_manager

    _add_sub_agents_postprocessor(
        postprocessor_manager=postprocessor_manager,
        tool_manager=tool_manager,
        config=config,
        response_watcher=common_components.response_watcher,
    )
    _add_sub_agents_evaluation(
        evaluation_manager=common_components.evaluation_manager,
        tool_manager=tool_manager,
        config=config,
        event=event,
        response_watcher=common_components.response_watcher,
    )

    return UniqueAI(
        event=event,
        config=config,
        logger=logger,
        chat_service=common_components.chat_service,
        content_service=common_components.content_service,
        tool_manager=tool_manager,
        thinking_manager=common_components.thinking_manager,
        history_manager=common_components.history_manager,
        reference_manager=common_components.reference_manager,
        streaming_handler=common_components.chat_service,
        evaluation_manager=common_components.evaluation_manager,
        postprocessor_manager=postprocessor_manager,
        debug_info_manager=debug_info_manager,
        mcp_servers=event.payload.mcp_servers,
        message_step_logger=common_components.message_step_logger,
    )


def _add_sub_agents_postprocessor(
    postprocessor_manager: PostprocessorManager,
    tool_manager: ToolManager | ResponsesApiToolManager,
    config: UniqueAIConfig,
    response_watcher: SubAgentResponseWatcher,
) -> None:
    sub_agents = tool_manager.sub_agents
    if len(sub_agents) > 0:
        display_config = SubAgentResponsesPostprocessorConfig(
            sleep_time_before_update=config.agent.experimental.sub_agents_config.sleep_time_before_update,
        )
        display_specs = []
        for tool in sub_agents:
            tool_config = cast(
                ExtendedSubAgentToolConfig, tool.settings.configuration
            )  # (BeforeValidator of ToolBuildConfig)

            display_specs.append(
                SubAgentDisplaySpec(
                    assistant_id=tool_config.assistant_id,
                    display_name=tool.display_name(),
                    display_config=tool_config.response_display_config,
                )
            )
        reference_postprocessor = SubAgentReferencesPostprocessor(
            response_watcher=response_watcher,
        )
        sub_agent_responses_postprocessor = SubAgentResponsesDisplayPostprocessor(
            config=display_config,
            response_watcher=response_watcher,
            display_specs=display_specs,
        )
        postprocessor_manager.add_postprocessor(reference_postprocessor)
        postprocessor_manager.add_postprocessor(sub_agent_responses_postprocessor)


def _add_sub_agents_evaluation(
    evaluation_manager: EvaluationManager,
    tool_manager: ToolManager | ResponsesApiToolManager,
    config: UniqueAIConfig,
    event: ChatEvent,
    response_watcher: SubAgentResponseWatcher,
) -> None:
    sub_agents = tool_manager.sub_agents
    if (
        len(sub_agents) > 0
        and config.agent.experimental.sub_agents_config.evaluation_config is not None
    ):
        evaluation_specs = []
        for tool in sub_agents:
            tool_config = cast(
                ExtendedSubAgentToolConfig, tool.settings.configuration
            )  # (BeforeValidator of ToolBuildConfig)

            evaluation_specs.append(
                SubAgentEvaluationSpec(
                    assistant_id=tool_config.assistant_id,
                    display_name=tool.display_name(),
                    config=tool_config.evaluation_config,
                )
            )

        sub_agent_evaluation = SubAgentEvaluationService(
            config=config.agent.experimental.sub_agents_config.evaluation_config,
            language_model_service=LanguageModelService.from_event(event),
            evaluation_specs=evaluation_specs,
            response_watcher=response_watcher,
        )
        evaluation_manager.add_evaluation(sub_agent_evaluation)



from enum import StrEnum
from pathlib import Path
from typing import Annotated, Any, Generic, Literal, TypeVar

from pydantic import BaseModel, Field, ValidationInfo, field_validator, model_validator
from unique_deep_research.config import DeepResearchToolConfig
from unique_deep_research.service import DeepResearchTool
from unique_follow_up_questions.config import FollowUpQuestionsConfig
from unique_internal_search.config import InternalSearchConfig
from unique_internal_search.service import InternalSearchTool
from unique_stock_ticker.config import StockTickerConfig
from unique_swot import SwotAnalysisTool, SwotAnalysisToolConfig
from unique_toolkit._common.validators import (
    LMI,
    ClipInt,
    get_LMI_default_field,
)
from unique_toolkit.agentic.evaluation.hallucination.constants import (
    HallucinationConfig,
)
from unique_toolkit.agentic.evaluation.schemas import EvaluationMetricName
from unique_toolkit.agentic.history_manager.history_manager import (
    UploadedContentConfig,
)
from unique_toolkit.agentic.responses_api import (
    DisplayCodeInterpreterFilesPostProcessorConfig,
    ShowExecutedCodePostprocessorConfig,
)
from unique_toolkit.agentic.tools.a2a import (
    REFERENCING_INSTRUCTIONS_FOR_SYSTEM_PROMPT,
    REFERENCING_INSTRUCTIONS_FOR_USER_PROMPT,
)
from unique_toolkit.agentic.tools.a2a.evaluation import SubAgentEvaluationServiceConfig
from unique_toolkit.agentic.tools.config import get_configuration_dict
from unique_toolkit.agentic.tools.openai_builtin.manager import (
    OpenAICodeInterpreterConfig,
)
from unique_toolkit.agentic.tools.tool import ToolBuildConfig
from unique_toolkit.agentic.tools.tool_progress_reporter import (
    ToolProgressReporterConfig,
)
from unique_toolkit.language_model.default_language_model import DEFAULT_GPT_4o
from unique_web_search.config import WebSearchConfig
from unique_web_search.service import WebSearchTool

DeactivatedNone = Annotated[
    None,
    Field(title="Deactivated", description="None"),
]


class SpaceType(StrEnum):
    UNIQUE_CUSTOM = "unique_custom"
    UNIQUE_AI = "unique_ai"
    UNIQUE_TRANSLATION = "unique_translation"
    UNIQUE_MAGIC_TABLE = ""


T = TypeVar("T", bound=SpaceType)


class SpaceConfigBase(BaseModel, Generic[T]):
    """Base class for space configuration."""

    model_config = get_configuration_dict(frozen=True)
    type: T = Field(description="The type of the space.")

    project_name: str = Field(
        default="Unique AI",
        description="The project name as optained from spaces 2.0",
    )

    language_model: LMI = get_LMI_default_field(DEFAULT_GPT_4o)

    custom_instructions: str = Field(
        default="",
        description="A custom instruction provided by the system admin.",
    )

    tools: list[ToolBuildConfig] = Field(
        default=[
            ToolBuildConfig(
                name=InternalSearchTool.name,
                configuration=InternalSearchConfig(
                    exclude_uploaded_files=True,
                ),
            ),
            ToolBuildConfig(
                name=WebSearchTool.name,
                configuration=WebSearchConfig(),
            ),
            ToolBuildConfig(
                name=DeepResearchTool.name,
                configuration=DeepResearchToolConfig(),
            ),
            ToolBuildConfig(
                name=SwotAnalysisTool.name,
                configuration=SwotAnalysisToolConfig(),
            ),
        ],
    )

    @field_validator("tools", mode="after")
    @classmethod
    def set_input_context_size(
        cls, tools: list[ToolBuildConfig], info: ValidationInfo
    ) -> list[ToolBuildConfig]:
        for tool in tools:
            if tool.name == InternalSearchTool.name:
                tool.configuration.language_model_max_input_tokens = (  # type: ignore
                    info.data["language_model"].token_limits.token_limit_input
                )
            elif tool.name == WebSearchTool.name:
                tool.configuration.language_model_max_input_tokens = (  # type: ignore
                    info.data["language_model"].token_limits.token_limit_input
                )
        return tools


class UniqueAISpaceConfig(SpaceConfigBase):
    """Contains configuration for the entities that a space provides."""

    type: Literal[SpaceType.UNIQUE_AI] = SpaceType.UNIQUE_AI


UniqueAISpaceConfig.model_rebuild()

LIMIT_MAX_TOOL_CALLS_PER_ITERATION = 50


class LoopConfiguration(BaseModel):
    model_config = get_configuration_dict()

    max_tool_calls_per_iteration: Annotated[
        int,
        *ClipInt(min_value=1, max_value=LIMIT_MAX_TOOL_CALLS_PER_ITERATION),
    ] = 10


class EvaluationConfig(BaseModel):
    model_config = get_configuration_dict()
    max_review_steps: int = 3
    hallucination_config: HallucinationConfig = HallucinationConfig()


# ------------------------------------------------------------
# Space 2.0 Config
# ------------------------------------------------------------


class UniqueAIPromptConfig(BaseModel):
    model_config = get_configuration_dict(frozen=True)

    system_prompt_template: str = Field(
        default_factory=lambda: (
            Path(__file__).parent / "prompts" / "system_prompt.jinja2"
        ).read_text(),
        description="The system prompt template as a Jinja2 template string.",
    )

    user_message_prompt_template: str = Field(
        default_factory=lambda: (
            Path(__file__).parent / "prompts" / "user_message_prompt.jinja2"
        ).read_text(),
        description="The user message prompt template as a Jinja2 template string.",
    )

    user_metadata: list[str] = Field(
        default=[],
        title="User Metadata",
        description="User metadata fields to be ingested in the system prompt.",
    )


class UniqueAIServices(BaseModel):
    """Determine the services the agent is using

    All services are optional and can be disabled by setting them to None.
    """

    model_config = get_configuration_dict(frozen=True)

    follow_up_questions_config: (
        Annotated[
            FollowUpQuestionsConfig,
            Field(
                title="Active",
            ),
        ]
        | DeactivatedNone
    ) = FollowUpQuestionsConfig()

    stock_ticker_config: (
        Annotated[StockTickerConfig, Field(title="Active")] | DeactivatedNone
    ) = StockTickerConfig()

    evaluation_config: (
        Annotated[
            EvaluationConfig,
            Field(title="Active"),
        ]
        | DeactivatedNone
    ) = EvaluationConfig(
        hallucination_config=HallucinationConfig(),
        max_review_steps=0,
    )

    uploaded_content_config: UploadedContentConfig = UploadedContentConfig()

    tool_progress_reporter_config: ToolProgressReporterConfig = (
        ToolProgressReporterConfig()
    )


class InputTokenDistributionConfig(BaseModel):
    model_config = get_configuration_dict(frozen=True)

    percent_for_history: float = Field(
        default=0.2,
        ge=0.0,
        lt=1.0,
        description="The fraction of the max input tokens that will be reserved for the history.",
    )

    def max_history_tokens(self, max_input_token: int) -> int:
        return int(self.percent_for_history * max_input_token)


class SubAgentsReferencingConfig(BaseModel):
    model_config = get_configuration_dict()

    referencing_instructions_for_system_prompt: str = Field(
        default=REFERENCING_INSTRUCTIONS_FOR_SYSTEM_PROMPT,
        description="Referencing instructions for the main agent's system prompt.",
    )
    referencing_instructions_for_user_prompt: str = Field(
        default=REFERENCING_INSTRUCTIONS_FOR_USER_PROMPT,
        description="Referencing instructions for the main agent's user prompt. Should correspond to a short reminder.",
    )


class SubAgentsConfig(BaseModel):
    model_config = get_configuration_dict()

    referencing_config: (
        Annotated[SubAgentsReferencingConfig, Field(title="Active")] | DeactivatedNone
    ) = SubAgentsReferencingConfig()
    evaluation_config: (
        Annotated[SubAgentEvaluationServiceConfig, Field(title="Active")]
        | DeactivatedNone
    ) = SubAgentEvaluationServiceConfig()

    sleep_time_before_update: float = Field(
        default=0.5,
        description="Time to sleep before updating the main agent message to display the sub agent responses. Temporary fix to avoid rendering issues.",
    )


class CodeInterpreterExtendedConfig(BaseModel):
    model_config = get_configuration_dict()

    generated_files_config: DisplayCodeInterpreterFilesPostProcessorConfig = Field(
        default=DisplayCodeInterpreterFilesPostProcessorConfig(),
        title="Generated files config",
        description="Display config for generated files",
    )

    executed_code_display_config: (
        Annotated[
            ShowExecutedCodePostprocessorConfig,
            Field(title="Active"),
        ]
        | DeactivatedNone
    ) = Field(
        ShowExecutedCodePostprocessorConfig(),
        description="If active, generated code will be prepended to the LLM answer",
    )

    tool_config: OpenAICodeInterpreterConfig = Field(
        default=OpenAICodeInterpreterConfig(),
        title="Tool config",
    )


class ResponsesApiConfig(BaseModel):
    model_config = get_configuration_dict(frozen=True)

    code_interpreter: (
        Annotated[CodeInterpreterExtendedConfig, Field(title="Active")]
        | DeactivatedNone
    ) = Field(
        default=None,
        description="If active, the main agent will have acces to the OpenAI Code Interpreter tool",
    )

    use_responses_api: bool = Field(
        default=False,
        description="If set, the main agent will use the Responses API from OpenAI",
    )


class ExperimentalConfig(BaseModel):
    """Experimental features this part of the configuration might evolve in the future continuously"""

    model_config = get_configuration_dict(frozen=True)

    thinking_steps_display: bool = False

    # TODO: @gustavhartz, the Hallucination check should be triggered if enabled and the answer contains references.
    force_checks_on_stream_response_references: list[EvaluationMetricName] = Field(
        default=[EvaluationMetricName.HALLUCINATION],
        description="A list of checks to force on references. This is used to add hallucination check to references without new tool calls.",
    )

    # TODO: The temperature should be used via the additional_llm_options
    # then the additional_llm_options migth should eventually be closer to the LangaugeModelInfo
    temperature: float = Field(
        default=0.0,
        ge=0.0,
        le=10.0,
        description="The temperature to use for the LLM.",
    )

    additional_llm_options: dict[str, Any] = Field(
        default={},
        description="Additional options to pass to the LLM.",
    )

    loop_configuration: LoopConfiguration = LoopConfiguration(
        max_tool_calls_per_iteration=10
    )

    sub_agents_config: SubAgentsConfig = SubAgentsConfig()

    responses_api_config: ResponsesApiConfig = ResponsesApiConfig()


class UniqueAIAgentConfig(BaseModel):
    model_config = get_configuration_dict(frozen=True)

    max_loop_iterations: int = 8

    input_token_distribution: InputTokenDistributionConfig = Field(
        default=InputTokenDistributionConfig(),
        description="The distribution of the input tokens.",
    )

    prompt_config: UniqueAIPromptConfig = UniqueAIPromptConfig()

    services: UniqueAIServices = UniqueAIServices()

    experimental: ExperimentalConfig = ExperimentalConfig()


class UniqueAIConfig(BaseModel):
    model_config = get_configuration_dict(frozen=True)

    space: UniqueAISpaceConfig = UniqueAISpaceConfig()

    agent: UniqueAIAgentConfig = UniqueAIAgentConfig()

    @model_validator(mode="after")
    def disable_sub_agent_referencing_if_not_used(self) -> "UniqueAIConfig":
        if not any(tool.is_sub_agent for tool in self.space.tools):
            self.agent.experimental.sub_agents_config.referencing_config = None
        return self



from unittest.mock import MagicMock

import pytest


class TestGetFilteredUserMetadata:
    """Test suite for UniqueAI._get_filtered_user_metadata method"""

    @pytest.fixture
    def mock_unique_ai(self):
        """Create a minimal UniqueAI instance with mocked dependencies"""
        # Lazy import to avoid heavy dependencies at module import time
        from unique_orchestrator.unique_ai import UniqueAI

        mock_logger = MagicMock()

        # Create minimal event structure
        dummy_event = MagicMock()
        dummy_event.payload.assistant_message.id = "assist_1"
        dummy_event.payload.user_message.text = "query"

        # Create minimal config structure
        mock_config = MagicMock()
        mock_config.agent.prompt_config.user_metadata = []

        # Create minimal required dependencies
        mock_chat_service = MagicMock()
        mock_content_service = MagicMock()
        mock_debug_info_manager = MagicMock()
        mock_reference_manager = MagicMock()
        mock_thinking_manager = MagicMock()
        mock_tool_manager = MagicMock()
        mock_history_manager = MagicMock()
        mock_evaluation_manager = MagicMock()
        mock_postprocessor_manager = MagicMock()
        mock_streaming_handler = MagicMock()
        mock_message_step_logger = MagicMock()

        # Instantiate UniqueAI
        ua = UniqueAI(
            logger=mock_logger,
            event=dummy_event,
            config=mock_config,
            chat_service=mock_chat_service,
            content_service=mock_content_service,
            debug_info_manager=mock_debug_info_manager,
            streaming_handler=mock_streaming_handler,
            reference_manager=mock_reference_manager,
            thinking_manager=mock_thinking_manager,
            tool_manager=mock_tool_manager,
            history_manager=mock_history_manager,
            evaluation_manager=mock_evaluation_manager,
            postprocessor_manager=mock_postprocessor_manager,
            message_step_logger=mock_message_step_logger,
            mcp_servers=[],
        )

        return ua

    def test_returns_empty_dict_when_config_is_empty_list(self, mock_unique_ai):
        """Test that empty dict is returned when config.user_metadata is an empty list"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = []
        mock_unique_ai._event.payload.user_metadata = {
            "department": "Engineering",
            "role": "Developer",
        }

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {}
        assert isinstance(result, dict)

    def test_returns_empty_dict_when_user_metadata_is_none(self, mock_unique_ai):
        """Test that empty dict is returned when user_metadata is None"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = [
            "department",
            "role",
        ]
        mock_unique_ai._event.payload.user_metadata = None

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {}
        assert isinstance(result, dict)

    def test_returns_empty_dict_when_both_config_and_metadata_are_empty(
        self, mock_unique_ai
    ):
        """Test that empty dict is returned when both config and user_metadata are empty/None"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = []
        mock_unique_ai._event.payload.user_metadata = None

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {}
        assert isinstance(result, dict)

    def test_filters_metadata_to_include_only_configured_keys(self, mock_unique_ai):
        """Test that only keys specified in config are included in the result"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = [
            "department",
            "role",
        ]
        mock_unique_ai._event.payload.user_metadata = {
            "department": "Engineering",
            "role": "Developer",
            "location": "San Francisco",
            "salary": "100000",
        }

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {"department": "Engineering", "role": "Developer"}
        assert "location" not in result
        assert "salary" not in result
        # Verify all values are strings
        assert all(isinstance(v, str) for v in result.values())

    def test_returns_only_existing_keys_from_user_metadata(self, mock_unique_ai):
        """Test that keys in config but not in user_metadata are not included"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = [
            "department",
            "role",
            "team",
            "manager",
        ]
        mock_unique_ai._event.payload.user_metadata = {
            "department": "Engineering",
            "role": "Developer",
        }

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {"department": "Engineering", "role": "Developer"}
        assert "team" not in result
        assert "manager" not in result

    def test_handles_single_key_in_config(self, mock_unique_ai):
        """Test filtering with a single key in config"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = ["department"]
        mock_unique_ai._event.payload.user_metadata = {
            "department": "Engineering",
            "role": "Developer",
        }

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {"department": "Engineering"}
        assert isinstance(result["department"], str)

    def test_handles_string_values(self, mock_unique_ai):
        """Test that string values in user_metadata are preserved"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = [
            "name",
            "email",
            "department",
            "title",
        ]
        mock_unique_ai._event.payload.user_metadata = {
            "name": "John Doe",
            "email": "john.doe@example.com",
            "department": "Engineering",
            "title": "Senior Developer",
            "ignored": "This should not appear",
        }

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {
            "name": "John Doe",
            "email": "john.doe@example.com",
            "department": "Engineering",
            "title": "Senior Developer",
        }
        assert "ignored" not in result
        # Verify all values are strings
        assert all(isinstance(v, str) for v in result.values())

    def test_handles_empty_dict_user_metadata(self, mock_unique_ai):
        """Test behavior when user_metadata is an empty dict"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = ["department"]
        mock_unique_ai._event.payload.user_metadata = {}

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {}

    def test_handles_empty_string_values(self, mock_unique_ai):
        """Test that empty string values in user_metadata are preserved if key is in config"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = [
            "department",
            "role",
        ]
        mock_unique_ai._event.payload.user_metadata = {
            "department": "Engineering",
            "role": "",
        }

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {"department": "Engineering", "role": ""}
        assert isinstance(result["role"], str)

    def test_preserves_original_metadata_unchanged(self, mock_unique_ai):
        """Test that the original user_metadata dict is not modified"""
        original_metadata = {
            "department": "Engineering",
            "role": "Developer",
            "location": "San Francisco",
        }
        mock_unique_ai._config.agent.prompt_config.user_metadata = ["department"]
        mock_unique_ai._event.payload.user_metadata = original_metadata.copy()

        result = mock_unique_ai._get_filtered_user_metadata()

        # Original should still have all keys
        assert mock_unique_ai._event.payload.user_metadata == original_metadata
        # Result should only have filtered key
        assert result == {"department": "Engineering"}

    def test_handles_special_characters_in_values(self, mock_unique_ai):
        """Test that special characters in string values are preserved"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = [
            "description",
            "notes",
        ]
        mock_unique_ai._event.payload.user_metadata = {
            "description": "User with special chars: @#$%^&*()",
            "notes": "Multi-line\ntext\twith\ttabs",
            "other": "excluded",
        }

        result = mock_unique_ai._get_filtered_user_metadata()

        assert result == {
            "description": "User with special chars: @#$%^&*()",
            "notes": "Multi-line\ntext\twith\ttabs",
        }
        assert all(isinstance(v, str) for v in result.values())

    def test_return_type_is_dict_str_str(self, mock_unique_ai):
        """Test that return type is dict[str, str]"""
        mock_unique_ai._config.agent.prompt_config.user_metadata = [
            "department",
            "role",
        ]
        mock_unique_ai._event.payload.user_metadata = {
            "department": "Engineering",
            "role": "Developer",
        }

        result = mock_unique_ai._get_filtered_user_metadata()

        # Check it's a dict
        assert isinstance(result, dict)
        # Check all keys are strings
        assert all(isinstance(k, str) for k in result.keys())
        # Check all values are strings
        assert all(isinstance(v, str) for v in result.values())



from unittest.mock import AsyncMock, MagicMock

import pytest


class TestUpdateDebugInfoForToolControl:
    """Test suite for UniqueAI._update_debug_info_if_tool_took_control method"""

    @pytest.fixture
    def mock_unique_ai(self):
        """Create a minimal UniqueAI instance with mocked dependencies"""
        # Lazy import to avoid heavy dependencies at module import time
        from unique_orchestrator.unique_ai import UniqueAI

        mock_logger = MagicMock()

        # Create minimal event structure
        dummy_event = MagicMock()
        dummy_event.payload.assistant_message.id = "assist_1"
        dummy_event.payload.user_message.text = "query"
        dummy_event.payload.assistant_id = "assistant_123"
        dummy_event.payload.name = "TestAssistant"
        dummy_event.payload.user_metadata = {"key": "value"}
        dummy_event.payload.tool_parameters = {"param": "value"}

        # Create minimal config structure
        mock_config = MagicMock()
        mock_config.agent.prompt_config.user_metadata = []

        # Create minimal required dependencies
        mock_chat_service = MagicMock()
        mock_chat_service.update_debug_info_async = AsyncMock()
        mock_content_service = MagicMock()
        mock_debug_info_manager = MagicMock()
        mock_reference_manager = MagicMock()
        mock_thinking_manager = MagicMock()
        mock_tool_manager = MagicMock()
        mock_history_manager = MagicMock()
        mock_evaluation_manager = MagicMock()
        mock_postprocessor_manager = MagicMock()
        mock_streaming_handler = MagicMock()
        mock_message_step_logger = MagicMock()

        # Instantiate UniqueAI
        ua = UniqueAI(
            logger=mock_logger,
            event=dummy_event,
            config=mock_config,
            chat_service=mock_chat_service,
            content_service=mock_content_service,
            debug_info_manager=mock_debug_info_manager,
            streaming_handler=mock_streaming_handler,
            reference_manager=mock_reference_manager,
            thinking_manager=mock_thinking_manager,
            tool_manager=mock_tool_manager,
            history_manager=mock_history_manager,
            evaluation_manager=mock_evaluation_manager,
            postprocessor_manager=mock_postprocessor_manager,
            message_step_logger=mock_message_step_logger,
            mcp_servers=[],
        )

        return ua

    @pytest.mark.ai
    @pytest.mark.asyncio
    async def test_does_not_update_debug_info_when_tool_did_not_take_control(
        self, mock_unique_ai
    ) -> None:
        """
        Purpose: Verify that _update_debug_info_if_tool_took_control does nothing when
        _tool_took_control is False.
        Why this matters: Debug info should only be updated when a tool takes control
        of the conversation. When no tool takes control, no update should occur.
        Setup summary: Create a UniqueAI instance with _tool_took_control set to False.
        """
        # Arrange
        mock_unique_ai._tool_took_control = False
        mock_unique_ai._debug_info_manager.get.return_value = {
            "tools": [{"name": "SearchTool"}]
        }

        # Act
        await mock_unique_ai._update_debug_info_if_tool_took_control()

        # Assert
        mock_unique_ai._chat_service.update_debug_info_async.assert_not_called()

    @pytest.mark.ai
    @pytest.mark.asyncio
    async def test_does_not_update_debug_info_if_tool_took_control_and_deep_research_is_in_tool_names(
        self, mock_unique_ai
    ) -> None:
        """
        Purpose: Verify that _update_debug_info_if_tool_took_control does nothing when
        DeepResearch is among the called tools.
        Why this matters: DeepResearch handles debug info directly since it calls
        the orchestrator multiple times, so we should not update debug info here.
        Setup summary: Create a UniqueAI instance with _tool_took_control set to True
        and DeepResearch in the tools list.
        """
        # Arrange
        mock_unique_ai._tool_took_control = True
        mock_unique_ai._debug_info_manager.get.return_value = {
            "tools": [{"name": "DeepResearch"}, {"name": "SearchTool"}]
        }

        # Act
        await mock_unique_ai._update_debug_info_if_tool_took_control()

        # Assert
        mock_unique_ai._chat_service.update_debug_info_async.assert_not_called()

    @pytest.mark.ai
    @pytest.mark.asyncio
    async def test_does_not_update_debug_info_if_tool_took_control_and_only_deep_research_is_called(
        self, mock_unique_ai
    ) -> None:
        """
        Purpose: Verify that _update_debug_info_if_tool_took_control does nothing when
        only DeepResearch is the called tool.
        Why this matters: Even when DeepResearch is the only tool, we should still
        skip the debug info update because DeepResearch manages its own debug info.
        Setup summary: Create a UniqueAI instance with _tool_took_control set to True
        and only DeepResearch in the tools list.
        """
        # Arrange
        mock_unique_ai._tool_took_control = True
        mock_unique_ai._debug_info_manager.get.return_value = {
            "tools": [{"name": "DeepResearch"}]
        }

        # Act
        await mock_unique_ai._update_debug_info_if_tool_took_control()

        # Assert
        mock_unique_ai._chat_service.update_debug_info_async.assert_not_called()

    @pytest.mark.ai
    @pytest.mark.asyncio
    async def test_updates_debug_info_when_tool_took_control_without_deep_research(
        self, mock_unique_ai
    ) -> None:
        """
        Purpose: Verify that _update_debug_info_if_tool_took_control updates debug info
        when a tool takes control and DeepResearch is not among the tools.
        Why this matters: When a non-DeepResearch tool takes control, we need to
        update the debug info with the relevant conversation context.
        Setup summary: Create a UniqueAI instance with _tool_took_control set to True
        and no DeepResearch in the tools list.
        """
        # Arrange
        mock_unique_ai._tool_took_control = True
        debug_info = {"tools": [{"name": "SearchTool"}, {"name": "WebSearch"}]}
        mock_unique_ai._debug_info_manager.get.return_value = debug_info

        # Act
        await mock_unique_ai._update_debug_info_if_tool_took_control()

        # Assert
        expected_debug_info_event = {
            "tools": debug_info,
            "assistant": {
                "id": "assistant_123",
                "name": "TestAssistant",
            },
            "chosenModule": "TestAssistant",
            "userMetadata": {"key": "value"},
            "toolParameters": {"param": "value"},
        }
        mock_unique_ai._chat_service.update_debug_info_async.assert_called_once_with(
            debug_info=expected_debug_info_event
        )

    @pytest.mark.ai
    @pytest.mark.asyncio
    async def test_updates_debug_info_with_single_tool(self, mock_unique_ai) -> None:
        """
        Purpose: Verify that _update_debug_info_if_tool_took_control correctly formats
        the debug info event with a single tool.
        Why this matters: The debug info event structure must be correct for the
        frontend to properly display the tool execution context.
        Setup summary: Create a UniqueAI instance with _tool_took_control set to True
        and a single non-DeepResearch tool.
        """
        # Arrange
        mock_unique_ai._tool_took_control = True
        debug_info = {"tools": [{"name": "SWOT"}]}
        mock_unique_ai._debug_info_manager.get.return_value = debug_info

        # Act
        await mock_unique_ai._update_debug_info_if_tool_took_control()

        # Assert
        expected_debug_info_event = {
            "tools": debug_info,
            "assistant": {
                "id": "assistant_123",
                "name": "TestAssistant",
            },
            "chosenModule": "TestAssistant",
            "userMetadata": {"key": "value"},
            "toolParameters": {"param": "value"},
        }
        mock_unique_ai._chat_service.update_debug_info_async.assert_called_once_with(
            debug_info=expected_debug_info_event
        )

    @pytest.mark.ai
    @pytest.mark.asyncio
    async def test_updates_debug_info_with_none_user_metadata(
        self, mock_unique_ai
    ) -> None:
        """
        Purpose: Verify that _update_debug_info_if_tool_took_control handles None
        user_metadata correctly.
        Why this matters: User metadata can be None, and the function should handle
        this gracefully without errors.
        Setup summary: Create a UniqueAI instance with _tool_took_control set to True,
        no DeepResearch in tools, and None user_metadata.
        """
        # Arrange
        mock_unique_ai._tool_took_control = True
        mock_unique_ai._event.payload.user_metadata = None
        debug_info = {"tools": [{"name": "SearchTool"}]}
        mock_unique_ai._debug_info_manager.get.return_value = debug_info

        # Act
        await mock_unique_ai._update_debug_info_if_tool_took_control()

        # Assert
        expected_debug_info_event = {
            "tools": debug_info,
            "assistant": {
                "id": "assistant_123",
                "name": "TestAssistant",
            },
            "chosenModule": "TestAssistant",
            "userMetadata": None,
            "toolParameters": {"param": "value"},
        }
        mock_unique_ai._chat_service.update_debug_info_async.assert_called_once_with(
            debug_info=expected_debug_info_event
        )

    @pytest.mark.ai
    @pytest.mark.asyncio
    async def test_updates_debug_info_with_none_tool_parameters(
        self, mock_unique_ai
    ) -> None:
        """
        Purpose: Verify that _update_debug_info_if_tool_took_control handles None
        tool_parameters correctly.
        Why this matters: Tool parameters can be None, and the function should handle
        this gracefully without errors.
        Setup summary: Create a UniqueAI instance with _tool_took_control set to True,
        no DeepResearch in tools, and None tool_parameters.
        """
        # Arrange
        mock_unique_ai._tool_took_control = True
        mock_unique_ai._event.payload.tool_parameters = None
        debug_info = {"tools": [{"name": "SearchTool"}]}
        mock_unique_ai._debug_info_manager.get.return_value = debug_info

        # Act
        await mock_unique_ai._update_debug_info_if_tool_took_control()

        # Assert
        expected_debug_info_event = {
            "tools": debug_info,
            "assistant": {
                "id": "assistant_123",
                "name": "TestAssistant",
            },
            "chosenModule": "TestAssistant",
            "userMetadata": {"key": "value"},
            "toolParameters": None,
        }
        mock_unique_ai._chat_service.update_debug_info_async.assert_called_once_with(
            debug_info=expected_debug_info_event
        )

    @pytest.mark.ai
    @pytest.mark.asyncio
    async def test_updates_debug_info_with_empty_tools_list(
        self, mock_unique_ai
    ) -> None:
        """
        Purpose: Verify that _update_debug_info_if_tool_took_control updates debug info
        when the tools list is empty.
        Why this matters: Even with an empty tools list, if _tool_took_control is True,
        the debug info should still be updated (edge case).
        Setup summary: Create a UniqueAI instance with _tool_took_control set to True
        and an empty tools list.
        """
        # Arrange
        mock_unique_ai._tool_took_control = True
        debug_info = {"tools": []}
        mock_unique_ai._debug_info_manager.get.return_value = debug_info

        # Act
        await mock_unique_ai._update_debug_info_if_tool_took_control()

        # Assert
        expected_debug_info_event = {
            "tools": debug_info,
            "assistant": {
                "id": "assistant_123",
                "name": "TestAssistant",
            },
            "chosenModule": "TestAssistant",
            "userMetadata": {"key": "value"},
            "toolParameters": {"param": "value"},
        }
        mock_unique_ai._chat_service.update_debug_info_async.assert_called_once_with(
            debug_info=expected_debug_info_event
        )

    @pytest.mark.ai
    @pytest.mark.asyncio
    async def test_deep_research_check_is_case_sensitive(self, mock_unique_ai) -> None:
        """
        Purpose: Verify that the DeepResearch check is case-sensitive.
        Why this matters: The check for "DeepResearch" should be exact match.
        Tools with different casing (e.g., "deepresearch", "DEEPRESEARCH") should
        not trigger the early return.
        Setup summary: Create a UniqueAI instance with _tool_took_control set to True
        and tools with similar but not exact "DeepResearch" names.
        """
        # Arrange
        mock_unique_ai._tool_took_control = True
        debug_info = {"tools": [{"name": "deepresearch"}, {"name": "DEEPRESEARCH"}]}
        mock_unique_ai._debug_info_manager.get.return_value = debug_info

        # Act
        await mock_unique_ai._update_debug_info_if_tool_took_control()

        # Assert - should update because "DeepResearch" exact match is not found
        mock_unique_ai._chat_service.update_debug_info_async.assert_called_once()
